from dataclasses import dataclass
from enum import Enum

from mm_std import print_plain

from mm_btc.wallet import AddressType, derive_accounts, generate_mnemonic


class PrivateType(str, Enum):
    hex = "hex"
    wif = "wif"


@dataclass
class Args:
    mnemonic: str
    passphrase: str
    words: int
    limit: int
    hex: bool  # Print private key in hex format instead of WIF
    address_type: AddressType
    path: str
    testnet: bool


def run(args: Args) -> None:
    mnemonic = args.mnemonic or generate_mnemonic()
    passphrase = args.passphrase
    path = get_derivation_path_prefix(args.path, args.testnet)
    accounts = derive_accounts(mnemonic, passphrase, path, args.address_type, args.limit)

    print_plain(f"{mnemonic}")
    if passphrase:
        print_plain(f"{passphrase}")
    for acc in accounts:
        private = acc.private if args.hex else acc.wif
        print_plain(f"{acc.path} {acc.address} {private}")


def get_derivation_path_prefix(path: str, testnet: bool) -> str:
    if path.startswith("m/"):
        return path
    coin = "1" if testnet else "0"
    if path == "bip44":
        return f"m/44'/{coin}'/0'/0"
    if path == "bip84":
        return f"m/84'/{coin}'/0'/0"

    raise ValueError("Invalid path")



from dataclasses import dataclass
from enum import Enum, unique

from hdwallet import HDWallet
from hdwallet.cryptocurrencies import Bitcoin
from hdwallet.derivations import CustomDerivation
from hdwallet.hds import BIP32HD
from hdwallet.mnemonics import BIP39Mnemonic
from mnemonic import Mnemonic

BIP44_MAINNET_PATH = "m/44'/0'/0'/0"
BIP44_TESTNET_PATH = "m/44'/1'/0'/0"
BIP84_MAINNET_PATH = "m/84'/0'/0'/0"
BIP84_TESTNET_PATH = "m/84'/1'/0'/0"


@dataclass
class Account:
    address: str
    private: str
    wif: str
    path: str


@unique
class AddressType(str, Enum):
    P2PKH = "P2PKH"  # Pay to Public Key Hash
    P2SH = "P2SH"  # Pay to Script Hash
    P2TR = "P2TR"  # Taproot
    P2WPKH = "P2WPKH"  # Native SegWit
    P2WPKH_IN_P2SH = "P2WPKH-In-P2SH"
    P2WSH = "P2WSH"  # Native SegWit
    P2WSH_IN_P2SH = "P2WSH-In-P2SH"


def generate_mnemonic(language: str = "english", words: int = 12) -> str:
    return Mnemonic(language).generate(strength=mnemonic_words_to_strenght(words))


def derive_accounts(mnemonic: str, passphrase: str, path_prefix: str, address_type: AddressType, limit: int) -> list[Account]:
    coin = Bitcoin
    if path_prefix.startswith(("m/84'/1'", "m/44'/1'")):
        network = coin.NETWORKS.TESTNET
    elif path_prefix.startswith(("m/84'/0'", "m/44'/0'")):
        network = coin.NETWORKS.MAINNET
    else:
        raise ValueError("Invalid path")

    wallet = HDWallet(cryptocurrency=coin, network=network, hd=BIP32HD, passphrase=passphrase).from_mnemonic(
        BIP39Mnemonic(mnemonic)
    )
    wallet.clean_derivation()

    accounts = []
    for index_path in range(limit):
        wallet.clean_derivation()
        path = f"{path_prefix}/{index_path}"
        w = wallet.from_derivation(derivation=CustomDerivation(path=path))
        accounts.append(Account(address=w.address(address_type), private=w.private_key(), wif=w.wif(), path=path))
        w.clean_derivation()

    return accounts


def mnemonic_words_to_strenght(words: int) -> int:
    if words == 12:
        return 128
    if words == 15:
        return 160
    if words == 18:
        return 192
    if words == 21:
        return 224
    if words == 24:
        return 256

    raise ValueError("Invalid words")


def is_testnet_address(address: str) -> bool:
    return address.startswith(("m", "n", "tb1"))



from mm_std import print_json

from mm_btc.blockstream import BlockstreamClient
from mm_btc.wallet import is_testnet_address


async def run(address: str) -> None:
    client = BlockstreamClient(testnet=is_testnet_address(address))
    res = await client.get_utxo(address)
    print_json(res.value_or_error())



from collections.abc import Sequence

from mm_std import HttpResponse, Result, http_request, random_str_choice
from pydantic import BaseModel

MAINNET_BASE_URL = "https://blockstream.info/api"
TESTNET_BASE_URL = "https://blockstream.info/testnet/api"

# ERROR_INVALID_ADDRESS = "INVALID_ADDRESS"
# ERROR_INVALID_NETWORK = "INVALID_NETWORK"

ERROR_400_BAD_REQUEST = "400 Bad Request"

type Proxy = str | Sequence[str] | None


class Mempool(BaseModel):
    count: int
    vsize: int
    total_fee: int
    fee_histogram: list[tuple[float, int]]


class Address(BaseModel):
    class ChainStats(BaseModel):
        funded_txo_count: int
        funded_txo_sum: int
        spent_txo_count: int
        spent_txo_sum: int
        tx_count: int

    class MempoolStats(BaseModel):
        funded_txo_count: int
        funded_txo_sum: int
        spent_txo_count: int
        spent_txo_sum: int
        tx_count: int

    chain_stats: ChainStats
    mempool_stats: MempoolStats


class Utxo(BaseModel):
    class Status(BaseModel):
        confirmed: bool
        block_height: int
        block_hash: str
        block_time: int

    txid: str
    vout: int
    status: Status
    value: int


class BlockstreamClient:
    def __init__(self, testnet: bool = False, timeout: float = 5, proxies: Proxy = None, attempts: int = 1) -> None:
        self.testnet = testnet
        self.timeout = timeout
        self.proxies = proxies
        self.attempts = attempts
        self.base_url = TESTNET_BASE_URL if testnet else MAINNET_BASE_URL

    async def get_address(self, address: str) -> Result[Address]:
        result: Result[Address] = Result.err("not started yet")
        for _ in range(self.attempts):
            res = await self._request(f"/address/{address}")
            try:
                if res.status_code == 400:
                    return res.to_err("400 Bad Request")
                return res.to_ok(Address(**res.parse_json_body()))
            except Exception as e:
                result = res.to_err(e)
        return result

    async def get_confirmed_balance(self, address: str) -> Result[int]:
        return (await self.get_address(address)).and_then(
            lambda a: Result.ok(a.chain_stats.funded_txo_sum - a.chain_stats.spent_txo_sum)
        )

    async def get_utxo(self, address: str) -> Result[list[Utxo]]:
        result: Result[list[Utxo]] = Result.err("not started yet")
        for _ in range(self.attempts):
            res = await self._request(f"/address/{address}/utxo")
            try:
                return res.to_ok([Utxo(**out) for out in res.parse_json_body()])
            except Exception as e:
                result = res.to_err(e)
        return result

    async def get_mempool(self) -> Result[Mempool]:
        result: Result[Mempool] = Result.err("not started yet")
        for _ in range(self.attempts):
            res = await self._request("/mempool")
            try:
                return res.to_ok(Mempool(**res.parse_json_body()))
            except Exception as e:
                result = res.to_err(e)
        return result

    async def _request(self, url: str) -> HttpResponse:
        return await http_request(f"{self.base_url}{url}", timeout=self.timeout, proxy=random_str_choice(self.proxies))



import pytest

from mm_btc.cli.cli import app
from mm_btc.cli.cmd.mnemonic_cmd import get_derivation_path_prefix


def test_get_derivation_path_prefix():
    assert get_derivation_path_prefix("m/11'/0'/0'/0", testnet=True) == "m/11'/0'/0'/0"
    assert get_derivation_path_prefix("bip44", False) == "m/44'/0'/0'/0"
    assert get_derivation_path_prefix("bip44", True) == "m/44'/1'/0'/0"
    assert get_derivation_path_prefix("bip84", False) == "m/84'/0'/0'/0"
    assert get_derivation_path_prefix("bip84", True) == "m/84'/1'/0'/0"

    with pytest.raises(ValueError):
        get_derivation_path_prefix("bip", True)


def test_mnemonic_cmd(cli_runner, mnemonic, passphrase):
    cmd = f"mnemonic -m '{mnemonic}' --passphrase '{passphrase}' -a P2WPKH"
    res = cli_runner.invoke(app, cmd)
    assert res.exit_code == 0
    acc_line = "m/44'/0'/0'/0/7 bc1qvcq77599py2x46weksqf9zvf7a53aw9wefrrhs L3QF5FHUtX2a1ucGgVfrdhdvLHBSxsRQoGsv7tyY8P5Jt7UV9LZv"
    assert acc_line in res.stdout



import asyncio
import importlib.metadata
from pathlib import Path
from typing import Annotated

import typer
import typer.core
from mm_std import print_plain

from mm_btc.wallet import AddressType

from .cmd import address_cmd, create_tx_cmd, decode_tx_cmd, mnemonic_cmd, utxo_cmd

app = typer.Typer(no_args_is_help=True, pretty_exceptions_enable=False, add_completion=False)


@app.command("mnemonic")
@app.command(name="m", hidden=True)
def mnemonic_command(  # nosec B107:hardcoded_password_default
    mnemonic: Annotated[str, typer.Option("--mnemonic", "-m", help="")] = "",
    passphrase: Annotated[str, typer.Option("--passphrase", "-p")] = "",
    path: Annotated[str, typer.Option("--path", help="Derivation path. Examples: bip44, bip88, m/44'/0'/0'/0")] = "bip44",
    address_type: Annotated[AddressType, typer.Option("--address-type", "-a", help="Bitcoin address type")] = AddressType.P2WPKH,
    hex_: Annotated[bool, typer.Option("--hex", help="Print private key in hex format instead of WIF")] = False,
    words: int = typer.Option(12, "--words", "-w", help="Number of mnemonic words"),
    limit: int = typer.Option(10, "--limit", "-l"),
    testnet: bool = typer.Option(False, "--testnet", "-t", help="Testnet network"),
) -> None:
    """Generate keys based on a mnemonic"""
    mnemonic_cmd.run(
        mnemonic_cmd.Args(
            mnemonic=mnemonic,
            passphrase=passphrase,
            words=words,
            limit=limit,
            path=path,
            address_type=address_type,
            hex=hex_,
            testnet=testnet,
        )
    )


@app.command(name="address")
@app.command(name="a", hidden=True)
def address_command(address: str) -> None:
    """Get address info from Blockstream API"""
    asyncio.run(address_cmd.run(address))


@app.command("create-tx")
def create_tx_command(config_path: Annotated[Path, typer.Argument(exists=True)]) -> None:
    """Create a transaction"""
    create_tx_cmd.run(config_path)


@app.command("decode-tx")
def decode_tx_command(tx_hex: str, testnet: Annotated[bool, typer.Option("--testnet", "-t")] = False) -> None:
    """Decode a transaction"""
    decode_tx_cmd.run(tx_hex, testnet)


@app.command("utxo")
def utxo_command(address: str) -> None:
    """Get UTXOs from Blockstream API"""
    asyncio.run(utxo_cmd.run(address))


def version_callback(value: bool) -> None:
    if value:
        print_plain(f"mm-btc: v{importlib.metadata.version('mm-btc')}")
        raise typer.Exit


@app.callback()
def main(_version: bool = typer.Option(None, "--version", callback=version_callback, is_eager=True)) -> None:
    pass


if __name__ == "__main_":
    app()
